import numpy as np
import pyintersection as pyi



def get_manifold_number(rpoint):
    return rpoint[1]

def get_rpoint_params(rpoint):
    return rpoint[2], rpoint[3]

def generate_new_points(rpoints, atol=None, agen=None, bgen=None):
    ares, bres = [], []
    for rpoint in rpoints:
        n, t = get_rpoint_params(rpoint)
        manifold = get_manifold_number(rpoint)
        
        if manifold == 0:
            new_points = agen(n, t, atol)
            ares.append(new_points)
        elif manifold == 1:
            new_points = bgen(n, t, atol)
            bres.append(new_points)
        else:
            raise Exception("Wrong manifold number")
          
    counter = 0
    for nparr in ares:
        for unarr in nparr:
            counter += 1
    ares = np.reshape(np.array(ares), (counter, 5))
            
    counter = 0
    for nparr in bres:
        for unarr in nparr:
            counter += 1
    bres = np.reshape(np.array(bres), (counter, 5))
        
    return ares, bres

def optimize3d(a, b, target_atol=None, agen=None, bgen=None, max_iter=8):
    """
    optimize3d(apoints, bpoints, target_atol=None,
            agen=None, bgen=None, max_iter=8)

    Find intersections between two manifolds.

    Parameters
    ----------
    apoints: (N, 5) numpy.ndarray
        First manifold.
    bpoints: (N, 5) numpy.ndarray
        Second manifold.
    target_atol: float > 0
        Target optimize tolerance.
    tree_depth: int > 0
        Depth of sieve tree.

    Returns
    -------
    rcubes: (N, 7) numpy.ndarray
        Intersection cubes (ndarray): [cid, x_down, x_up, y_down, y_up, z_down, z_up].
    points: (N, 7) numpy.ndarray
        Intersection points (numpy.ndarray): [cid, m, n, t, x, y, z].
    tols: (3,) numpy.ndarray
        Tolerance by all axis: [x_tol, y_tol, z_tol].

    Examples
    --------
    >>> import pyintersection as pyi
    >>> import numpy as np
    >>> apoints = np.array([[0, 0.0, 1, 1, 1], [1, 0.1, 1, 1, 2], [2, 0.2, 1, 1, 3]])
    >>> bpoints = np.array([[0, 0.3, 1, 2, 2], [1, 0.4, 1, 1, 2], [2, 0.5, 1, 0, 2]])
    >>> cubes, points, tols = pyi.intersect3d(apoints, bpoints, 0.5)
    >>> print(cubes)
        [[0.  1.  1.  1.  1.5 1.5 2. ]]
    >>> print(points)
        [[0.  0.  1.  0.1 1.  1.  2. ]
        [0.  1.  1.  0.1 1.  1.  2. ]]
    >>> print(tols)
        [0.  0.5 0.5]
    """
    atol = max(pyi.get_boundary_cube3d(a, b))
    for i in range(max_iter):
        rcubes, rpoints, atols = pyi.intersect3d(a, b, atol=atol/2)
        #print(f'Iteration {i}: target_atols={atol/2} real_atols={max(atols)} rpoins={len(rpoints)} cubes={len(rcubes)}')
        if max(atols) <= target_atol:
            break
        atol = max(atols)
        a, b = generate_new_points(rpoints, atol=atol, agen=agen, bgen=bgen)
    return rcubes, rpoints, atols


import numpy as np
from extmodule import __intersect3d, __intersect6d, __get_boundary_cube3d


def check_manifold_type(a, dims):
    if not isinstance(a, np.ndarray):
        raise Exception("Error: ")

    if a.dtype != 'float64':
        raise Exception("Error: type")

    if len(a.shape) != 2:
        raise Exception("Error: shape")

    if a.shape[1] != dims:
        raise Exception("Error: shape")


def check_tolerance(atol):
    if not isinstance(atol, float):
        raise Exception("Error: tolerance should has float type")

    if atol <= 0:
        raise Exception("Error: tolerance should be greater than 0")


def intersect3d(a, b, atol=None, tree_depth=2):
    """
    intersect3d(apoints, bpoints, atol=None, tree_depth=2)

    Find intersections between two manifolds.

    Parameters
    ----------
    apoints: (N, 5) numpy.ndarray
        First manifold.
    bpoints: (N, 5) numpy.ndarray
        Second manifold.
    atol: float > 0
        Target intersect tolerance.
    tree_depth: int > 0
        Depth of sieve tree.

    Returns
    -------
    rcubes: (N, 7) numpy.ndarray
        Intersection cubes (ndarray): [cid, x_down, x_up, y_down, y_up, z_down, z_up].
    points: (N, 7) numpy.ndarray
        Intersection points (numpy.ndarray): [cid, m, n, t, x, y, z].
    tols: (3,) numpy.ndarray
        Tolerance by all axis: [x_tol, y_tol, z_tol].

    Examples
    --------
    >>> import pyintersection as pyi
    >>> import numpy as np
    >>> apoints = np.array([[0, 0.0, 1, 1, 1], [1, 0.1, 1, 1, 2], [2, 0.2, 1, 1, 3]])
    >>> bpoints = np.array([[0, 0.3, 1, 2, 2], [1, 0.4, 1, 1, 2], [2, 0.5, 1, 0, 2]])
    >>> cubes, points, tols = pyi.intersect3d(apoints, bpoints, 0.5)
    >>> print(cubes)
        [[0.  1.  1.  1.  1.5 1.5 2. ]]
    >>> print(points)
        [[0.  0.  1.  0.1 1.  1.  2. ]
        [0.  1.  1.  0.1 1.  1.  2. ]]
    >>> print(tols)
        [0.  0.5 0.5]
    """
    check_manifold_type(a, 5)
    check_manifold_type(b, 5)
    check_tolerance(atol)
    if a.shape[0] == 0 or b.shape[0] == 0:
        return np.zeros((0, 7)), np.zeros((0, 7)), np.zeros((3))
    return __intersect3d(a, b, atol, tree_depth)


def intersect6d(a, b, atol=None, tree_depth=2):
    """
    intersect6d(apoints, bpoints, atol=None, tree_depth=2)

    Find intersections between two manifolds.

    Parameters
    ----------
    apoints: (N, 8) numpy.ndarray
        First manifold.
    bpoints: (N, 8) numpy.ndarray
        Second manifold.
    atol: float > 0
        Target intersect tolerance.
    tree_depth: int > 0
        Depth of sieve tree.

    Returns
    -------
    rcubes: (N, 13) numpy.ndarray
        Intersection cubes (numpy.ndarray): [cid, x_down, x_up, y_down, y_up, z_down, z_up,
        vx_down, vx_up, vy_down, vy_up, vz_down, vz_up].
    points: (N, 10) numpy.ndarray
        Intersection points (numpy.ndarray): [cid, m, n, t, x, y, z, vx, vy, vz].
    tols: (6,) numpy.ndarray
        Tolerance by all axis: [x_tol, y_tol, z_tol, vx_tol, vy_tol, vz_tol].

    Examples
    --------
    >>> import pyintersection as pyi
    >>> import numpy as np
    >>> cubes, points, tols = pyi.intersect6d(np.array([]), np.array([]), 0.5)
    """
    check_manifold_type(a, 8)
    check_manifold_type(b, 8)
    check_tolerance(atol)
    if a.shape[0] == 0 or b.shape[0] == 0:
        return np.zeros((0, 13)), np.zeros((0, 10)), np.zeros((6))
    return __intersect6d(a, b, atol, tree_depth)

def get_boundary_cube3d(a, b):
    """
    get_boundary_cube3d(apoints, bpoints)

    Find boundary cube for two manifolds.

    Parameters
    ----------
    apoints: (N, 5) numpy.ndarray
        First manifold.
    bpoints: (N, 5) numpy.ndarray
        Second manifold.

    Returns
    -------
    cube: (6,) numpy.ndarray
        Boundary cube (numpy.ndarray): [x_down, x_up, y_down, y_up, z_down, z_up].

    Examples
    --------
    """
    check_manifold_type(a, 5)
    check_manifold_type(b, 5)
    if a.shape[0] == 0 or b.shape[0] == 0:
        return np.zeros(6)
    return __get_boundary_cube3d(a, b)



# The script that will build the module
# from distutils.core import setup, Extension
from setuptools import setup, Extension, find_packages
import sysconfig
import sys
import os.path as path
import site

# Find path to numpy include folder
def get_numpy_include():
    numpy_postfix = path.normpath('numpy/core/include/numpy')
    user_sp = site.getusersitepackages()
    sys_sp = site.getsitepackages()
    for sp in [user_sp] + sys_sp:
        numpy_include = path.join(sp, numpy_postfix)
        if path.exists(numpy_include):
            return numpy_include
    raise Exception('Numpy include directory not found.')

# Find paths to Python.h dir and numpy include dir
def get_include_dirs():
    paths = sysconfig.get_paths()
    python_include = paths['include']
    numpy_include = get_numpy_include()
    return python_include, numpy_include

def get_extra_compile_args():
    return ['/std:c++17'] if sys.platform == 'win32' else ['-std=c++17']


PYTHON_INCLUDE, NUMPY_INCLUDE = get_include_dirs() # numpy.get_include()
EXTRA_COMPILE_ARGS = get_extra_compile_args()

# Add extension to module
module = Extension('extmodule',
                    sources = ['pyintersection/extmodule/main.cpp'],
                    include_dirs = [PYTHON_INCLUDE, NUMPY_INCLUDE],
                    extra_compile_args = EXTRA_COMPILE_ARGS)

# Make setup
setup(name = 'pyintersection',
        version = '1.0',
        description = 'Pyintersection package',
        packages = find_packages(),
        ext_modules = [module])
